local M = {}

-- Create a new animation instance.
function M.newAnimation(image, frameWidth, frameHeight)
    local a = {}
    a.image = image
    a.frameWidth = frameWidth
    a.frameHeight = frameHeight
    a.currentQuad = love.graphics.newQuad(0, 0, frameWidth, frameHeight, image)
    a.animations = {}

    a.currentAnim = nil
    a.currentFrame = 1

    return a
end

-- Add a set of frames to an existing animation instance based on indices passed in, at a rate defined by animSpeed.
-- The animation will cycle through frames specified at these indices when played.
function M.add(anim, name, indices, animSpeed)
    local a = {}
    a.name = name
    a.speed = animSpeed
    a.frameCount = #indices

    local imageWidth = anim.image:getWidth() / anim.frameWidth
    
    a.quads = {}
    for _, index in ipairs(indices) do
        local cellX = index % imageWidth
        local cellY = math.floor(index / imageWidth)

        table.insert(a.quads, love.graphics.newQuad(
            cellX * anim.frameWidth,
            cellY * anim.frameHeight,
            anim.frameWidth,
            anim.frameHeight,
            anim.image
        ))
    end
    
    anim.animations[name] = a
end

-- Play an animation by name. Has no effect if the specified animation is already playing.
-- Optionally reset the frame counter to start animation from the beginning, defaults to false.
function M.play(anim, name, shouldReset)
    if anim.currentAnim and anim.currentAnim.name == name then return end

    local reset = shouldReset or false

    assert(anim.animations[name], "Animation of name \"" .. name .. "\" not found.")
    anim.currentAnim = anim.animations[name]

    if reset then
        M.resetFrameCounter(anim)
    end
end

-- Resets the frame counter on the animation. This can be useful to do manually.
function M.resetFrameCounter(anim)
    anim.currentFrame = 1
end

function M.update(anim)
    assert(anim.currentAnim, "No animations are currently playing.")

    anim.currentFrame = (anim.currentFrame + love.timer.getDelta() * anim.currentAnim.speed)
    if anim.currentFrame > anim.currentAnim.frameCount + 1 then
        anim.currentFrame = 1
    end

    local currentFrameIndex = math.floor(anim.currentFrame)
    anim.currentQuad = anim.currentAnim.quads[currentFrameIndex]
end

return M