local tm = require("tilemap")
local g  = require("globals")
local inspect = require("inspect")

local M = {}

-- Create a dialogue instance.
-- You should only need one of these, which you show and hide when necessary.
function M.newDialogue()
    local d = {}
    d.panel = tm.getLayerByName(g.tilemap, "dialogue")
    d.visible = false
    d.text = ""
    d.currentText = ""
    d.currentTextIndex = 0
    d.textTimer = 0
    d.textFinished = false
    d.cursorWave = 0
    return d
end

-- Show the dialogue panel and start text scrolling. Does nothing if already shown.
function M.show(dlg, text, sign)
    if dlg.visible then return end

    g.dialogueActive = true
    dlg.visible = true

    dlg.text = text
    dlg.currentText = ""
    dlg.currentTextIndex = 0
    dlg.textTimer = love.timer.getTime()

    dlg.signRef = sign
end

-- Hide the dialogue panel and reset properties. Does nothing if already hidden.
function M.hide(dlg)
    if not dlg.visible then return end

    g.dialogueActive = false
    dlg.visible = false

    dlg.text = ""
    dlg.currentText = ""
    dlg.textTimer = 0
    dlg.textFinished = false

    if dlg.signRef and dlg.signRef.onDialogueHidden then
        dlg.signRef:onDialogueHidden()
    end
end

function M.advance(dlg)
    if not dlg.visible and dlg.textFinished then return end

    dlg.textFinished = true
    dlg.currentText = dlg.text

    if dlg.signRef and dlg.signRef.onTextFinished then
        dlg.signRef:onTextFinished()
    end
end

function M.update(dlg)
    if not dlg.visible then return end

    dlg.cursorWave = math.sin(love.timer.getTime() * 10) * 5

    if not dlg.textFinished then
        if love.timer.getTime() - dlg.textTimer > 0.02 then
            dlg.textTimer = love.timer.getTime()
            dlg.currentTextIndex = dlg.currentTextIndex + 1
            dlg.currentText = dlg.currentText .. string.sub(dlg.text, dlg.currentTextIndex, dlg.currentTextIndex)
            if dlg.currentTextIndex >= #dlg.text then
                dlg.textFinished = true

                if dlg.signRef and dlg.signRef.onTextFinished then
                    dlg.signRef:onTextFinished()
                end
            end
        end
    end
end

function M.draw(dlg, cursorQuad)
    if not dlg.visible then return end

    local boxWidth = dlg.panel.width * g.TILE_SIZE
    local boxHeight = dlg.panel.height * g.TILE_SIZE
    local x = (love.graphics.getWidth() / 2) - boxWidth / 2
    local y = love.graphics.getHeight() - (2 * g.TILE_SIZE) - 20
    tm.drawLayer(g.tilemap, dlg.panel, x, y)

    local padding = 20
    local textX = x + padding
    local textY = y + padding

    love.graphics.setColor(0, 0, 0, 1)
    -- boxWidth * 2 because limit is applied before scaling
    love.graphics.printf(dlg.currentText, textX, textY, (boxWidth - padding * 2) * 2, "left", 0, 0.5)
    love.graphics.setColor(1, 1, 1, 1)

    if dlg.textFinished then
        love.graphics.draw(g.IMAGE_SHEET, cursorQuad, x + boxWidth - 20, y + boxHeight - 40 + dlg.cursorWave, math.pi / 2)
    end
end

return M