local g = require("globals")
local tm = require("tilemap")
local am = require("animation")
local sm = require("sign")
local util = require("util")
local inspect = require("inspect")

local M = {}

local function distanceTowardsTarget(obj, target)
    local disX = target.x - obj.x
    local disY = target.y - obj.y
    return math.sqrt(disX * disX + disY * disY)
end

local function angleTowardsTarget(obj, target)
    return math.atan2(target.y - obj.y, target.x - obj.x)
end

-- Move obj towards target by speed.
local function moveTowardsTarget(obj, target, speed)
    if not obj.moving then return end

    local distance = distanceTowardsTarget(obj, target)
    local angle = angleTowardsTarget(obj, target)

    local realSpeed = math.min(distance, speed)
    obj.x = obj.x + math.cos(angle) * realSpeed
    obj.y = obj.y + math.sin(angle) * realSpeed
end

function M.newPlayer(x, y)
    local p = {}

    p.graphic = am.newAnimation(g.IMAGE_SHEET, g.TILE_SIZE, g.TILE_SIZE)
    am.add(p.graphic, "idleDown", {7, 10}, 2)
    am.add(p.graphic, "idleUp", {14, 17}, 2)
    am.add(p.graphic, "idleRight", {21, 24}, 2)
    am.add(p.graphic, "idleLeft", {28, 31}, 2)
    am.add(p.graphic, "walkDown", {8, 7, 9, 7}, 6)
    am.add(p.graphic, "walkUp", {15, 14, 16, 14}, 6)
    am.add(p.graphic, "walkRight", {22, 21, 23, 21}, 6)
    am.add(p.graphic, "walkLeft", {29, 28, 30, 28}, 6)
    am.play(p.graphic, "idleDown")

    p.x = x
    p.y = y
    p.angle = 90 -- down
    p.moving = false
    p.movingLastFrame = false
    p.moveTarget = {x = x, y = y}
    p.speed = 2
    p.facing = "down"
    return p
end

function M.update(player)
    player.movingLastFrame = player.moving

    -- distance check used in below functions
    local distance = distanceTowardsTarget(player, player.moveTarget)
    player.moving = math.min(distance, player.speed) > 1

    M.move(player)

    moveTowardsTarget(player, player.moveTarget, player.speed)

    am.update(player.graphic)
    M.animate(player)
end

function M.move(player)
    if g.paused or g.dialogueActive then return end

    -- don't try to move while already moving, will desync from grid
    local shouldMove = false
    if not player.moving then
        if love.keyboard.isDown("w", "up") then
            player.moveTarget = {x = player.x, y = player.y - g.TILE_SIZE}
            shouldMove = true
        elseif love.keyboard.isDown("a", "left") then
            player.moveTarget = {x = player.x - g.TILE_SIZE, y = player.y}
            shouldMove = true
        elseif love.keyboard.isDown("s", "down") then
            player.moveTarget = {x = player.x, y = player.y + g.TILE_SIZE}
            shouldMove = true
        elseif love.keyboard.isDown("d", "right") then
            player.moveTarget = {x = player.x + g.TILE_SIZE, y = player.y}
            shouldMove = true
        end

        if shouldMove then
            local angleRad = angleTowardsTarget(player, player.moveTarget)
            player.angle = util.rad2deg(angleRad)
            M.updateFacing(player)
        end
    end

    if shouldMove then
        -- if moveTarget is inside a collidable tile, set moveTarget back to the player position
        if M.shouldCollide(player) then
            player.moveTarget.x = player.x
            player.moveTarget.y = player.y
            player.moving = false
        else
            player.moving = true
        end
    end
end

function M.shouldCollide(player)
    return tm.isTileAtPixel(g.tilemap, tm.getCollidableLayers(g.tilemap), player.moveTarget.x, player.moveTarget.y) or M.scanInFront(player)
end

-- Set player.facing based on the player's current angle.
function M.updateFacing(player)
    if player.angle < -135 then
        player.facing = "left"
    elseif player.angle < -45 then
        player.facing = "up"
    elseif player.angle < 45 then
        player.facing = "right"
    elseif player.angle < 135 then
        player.facing = "down"
    else
        player.facing = "left"
    end
end

function M.animate(player)
    if player.moving then
        if player.facing == "down" then
            am.play(player.graphic, "walkDown")
        elseif player.facing == "right" then
            am.play(player.graphic, "walkRight")
        elseif player.facing == "left" then
            am.play(player.graphic, "walkLeft")
        elseif player.facing == "up" then
            am.play(player.graphic, "walkUp")
        end
    else
        if player.facing == "down" then
            am.play(player.graphic, "idleDown", true)
        elseif player.facing == "right" then
            am.play(player.graphic, "idleRight", true)
        elseif player.facing == "left" then
            am.play(player.graphic, "idleLeft", true)
        elseif player.facing == "up" then
            am.play(player.graphic, "idleUp", true)
        end
    end
end

function M.handleKeyPressed(player, key, scancode, isrepeat)
    local handled = false

    if not player.moving and not g.paused and not g.dialogueActive then
        if key == "space" then
            -- if a collision happens, sign will never be nil
            local collided, sign = M.scanInFront(player)
            if collided then
                sm.onInteract(sign, player)
                player.moving = false
                g.SFX_CLICK:play()
            end

            handled = true
        end
    end

    return handled
end

-- Returns whether or not there is an entity in front of the player, as well as the sign object if one was found.
function M.scanInFront(player)
    local cx = player.x + player.graphic.frameWidth / 2
    local cy = player.y + player.graphic.frameHeight / 2
    local radAngle = util.deg2rad(player.angle)
    local targetX = cx + math.cos(radAngle) * g.TILE_SIZE
    local targetY = cy + math.sin(radAngle) * g.TILE_SIZE

    local centerPoint = {x = targetX, y = targetY}
    local collided = false
    local collidedSign
    for _, sign in ipairs(g.signs) do
        local signBounds = {x = sign.x, y = sign.y, w = sign.graphic.frameWidth, h = sign.graphic.frameHeight}
        if util.pointInRectangle(centerPoint, signBounds) then
            collided = true
            collidedSign = sign
            break
        end
    end

    return collided, collidedSign
end

function M.draw(player)
    love.graphics.draw(player.graphic.image, player.graphic.currentQuad, player.x, player.y)
end

return M