local json = require("json")
local g = require("globals")
local inspect = require("inspect")

local M = {}

-- Clamp num between max and min.
local function clamp(num, max, min)
    return math.min(math.max(num, min), max)
end

-- Returns a bounding rectangle (x, y, w, h) that contains all tiles in a layer. This is measured in map coordinates, not pixels.
local function findBounds(layer)
    local xValues = {}
    local yValues = {}

    for _, tile in pairs(layer.tiles) do
        table.insert(xValues, tile.x)
        table.insert(yValues, tile.y)
    end

    table.sort(xValues)
    table.sort(yValues)

    local bounds = {
        x = xValues[1],
        y = yValues[1],
        w = xValues[#xValues] - xValues[1] + 1,
        h = yValues[#yValues] - yValues[1] + 1
    }

    return bounds
end

-- Load a map from a JSON file path and a corresponding sprite sheet path.
function M.newTilemapFromFile(jsonPath, imagePath)
    local map = {}

    local contents, _ = love.filesystem.read(jsonPath)
    if contents ~= nil then
        map = json.decode(contents)
    else
        print("Map data failed to be read.")
    end

    map.loveImage = love.graphics.newImage(imagePath)
    local sheetWidth = map.loveImage:getWidth() / map.tileSize
    local sheetHeight = map.loveImage:getHeight() / map.tileSize
    local quadAmount = sheetWidth * sheetHeight

    map.loveQuads = {}
    for i = 0, quadAmount - 1 do
        local cellX = i % sheetWidth
        local cellY = math.floor(i / sheetWidth)

        local quad = love.graphics.newQuad(
            cellX * map.tileSize,
            cellY * map.tileSize,
            map.tileSize,
            map.tileSize,
            map.loveImage
        )
        assert(quad, "There was a problem creating a quad of ID: " .. i)

        -- sprite fusion IDs are stored as strings
        map.loveQuads[tostring(i)] = quad
    end

    -- create a map of indices so you can get layers by name
    -- also mark collidable layers
    map.layerIndices = {}
    map.collidableLayers = {}
    for i, layer in ipairs(map.layers) do
        map.layerIndices[layer.name] = i

        if layer.collider then
            table.insert(map.collidableLayers, layer.name)
        end

        local bounds = findBounds(layer)
        layer.width = bounds.w
        layer.height = bounds.h
    end

    return map
end

-- Check if there is a tile at the specified position using the table of layer names that is passed in.
function M.isTileAtPixel(map, layerNames, x, y)
    local cellX = clamp(math.floor(x / g.TILE_SIZE), map.mapWidth - 1, 0)
    local cellY = clamp(math.floor(y / g.TILE_SIZE), map.mapHeight - 1, 0)

    for _, name in ipairs(layerNames) do
        local layer = M.getLayerByName(map, name)
        for _, tile in ipairs(layer.tiles) do
            if cellX == tile.x and cellY == tile.y then
                return true
            end
        end
    end

    return false
end

function M.getLayerByName(map, name)
    return map.layers[map.layerIndices[name]]
end

-- Get the layer names of each collidable layer and returns them as a table.
function M.getCollidableLayers(map)
    return map.collidableLayers
end

-- Draw a map that has been loaded.
function M.drawMap(map)
    for _, layer in ipairs(map.layers) do
        M.drawLayer(map, layer)
    end
end

-- Draw a single layer on a loaded map. Optionally provide an offset position to draw at.
function M.drawLayer(map, layer, offsetX, offsetY)
    local ox = offsetX or 0
    local oy = offsetY or 0

    for _, tile in ipairs(layer.tiles) do
        if map.loveQuads[tile.id] then
            love.graphics.draw(
                map.loveImage,
                map.loveQuads[tile.id],
                tile.x * map.tileSize + ox,
                tile.y * map.tileSize + oy
            )
        end
    end
end

return M
